"""Tests arrow keys on both command and edit mode"""


from .utils import EDITOR_PAGE


def test_dualmode_arrows(notebook_frontend):

    # Tests in command mode.
    # Setting up the cells to test the keys to move up.
    notebook_frontend.to_command_mode()
    [notebook_frontend.press("b", page=EDITOR_PAGE) for i in range(3)]

    # Use both "k" and up arrow keys to moving up and enter a value.
    # Once located on the top cell, use the up arrow keys to prove the top cell is still selected.
    notebook_frontend.press("k", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.press("2", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    notebook_frontend.press("ArrowUp", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.press("1", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    notebook_frontend.press("k", page=EDITOR_PAGE)
    notebook_frontend.press("ArrowUp", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.press("0", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    assert notebook_frontend.get_cells_contents() == ["0", "1", "2", ""]

    # Use the "k" key on the top cell as well
    notebook_frontend.press("k", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.type(" edit #1", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    assert notebook_frontend.get_cells_contents() == ["0 edit #1", "1", "2", ""]

    # Setting up the cells to test the keys to move down
    [notebook_frontend.press("j", page=EDITOR_PAGE) for i in range(3)]
    [notebook_frontend.press("a", page=EDITOR_PAGE) for i in range(2)]
    notebook_frontend.press("k", page=EDITOR_PAGE)

    # Use both "j" key and down arrow keys to moving down and enter a value.
    # Once located on the bottom cell, use the down arrow key to prove the bottom cell is still selected.
    notebook_frontend.press("ArrowDown", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.press("3", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    notebook_frontend.press("j", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.press("4", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    notebook_frontend.press("j", page=EDITOR_PAGE)
    notebook_frontend.press("ArrowDown", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.press("5", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    assert notebook_frontend.get_cells_contents() == ["0 edit #1", "1", "2", "3", "4", "5"]

    # Use the "j" key on the top cell as well
    notebook_frontend.press("j", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.type(" edit #1", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    assert notebook_frontend.get_cells_contents() == ["0 edit #1", "1", "2", "3", "4", "5 edit #1"]

    # On the bottom cell, use both left and right arrow keys to prove the bottom cell is still selected.
    notebook_frontend.press("ArrowLeft", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.type(", #2", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    assert notebook_frontend.get_cells_contents() == ["0 edit #1", "1", "2", "3", "4", "5 edit #1, #2"]
    notebook_frontend.press("ArrowRight", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)
    notebook_frontend.type(" and #3", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    assert notebook_frontend.get_cells_contents() == ["0 edit #1", "1", "2", "3", "4", "5 edit #1, #2 and #3"]

    # Tests in edit mode.
    # First, erase the previous content and then setup the cells to test the keys to move up.
    [notebook_frontend.locate(".fa-cut.fa", page=EDITOR_PAGE).click() for i in range(6)]
    [notebook_frontend.press("b", page=EDITOR_PAGE) for i in range(2)]
    notebook_frontend.press("a", page=EDITOR_PAGE)
    notebook_frontend.press("Enter", page=EDITOR_PAGE)

    # Use the up arrow key to move down and enter a value.
    # We will use the left arrow key to move one char to the left since moving up on last character only moves selector to the first one.
    # Once located on the top cell, use the up arrow key to prove the top cell is still selected.
    notebook_frontend.press("ArrowUp", page=EDITOR_PAGE)
    notebook_frontend.press("1", page=EDITOR_PAGE)
    notebook_frontend.press("ArrowLeft", page=EDITOR_PAGE)
    [notebook_frontend.press("ArrowUp", page=EDITOR_PAGE) for i in range(2)]
    notebook_frontend.press("0", page=EDITOR_PAGE)

    # Use the down arrow key to move down and enter a value.
    # We will use the right arrow key to move one char to the right since moving down puts selector to the last character.
    # Once located on the bottom cell, use the down arrow key to prove the bottom cell is still selected.
    notebook_frontend.press("ArrowDown", page=EDITOR_PAGE)
    notebook_frontend.press("ArrowRight", page=EDITOR_PAGE)
    notebook_frontend.press("ArrowDown", page=EDITOR_PAGE)
    notebook_frontend.press("2", page=EDITOR_PAGE)
    [notebook_frontend.press("ArrowDown", page=EDITOR_PAGE) for i in range(2)]
    notebook_frontend.press("3", page=EDITOR_PAGE)
    notebook_frontend.to_command_mode()
    assert notebook_frontend.get_cells_contents() == ["0", "1", "2", "3"]



"""Test basic cell execution methods, related shortcuts, and error modes

Run this manually:
    # Normal pytest run
    pytest nbclassic/tests/end_to_end/test_interrupt.py
    # with playwright debug (run and poke around in the web console)
    PWDEBUG=1 pytest -s nbclassic/tests/end_to_end/test_interrupt.py
"""


from .utils import TREE_PAGE, EDITOR_PAGE


# # Use/uncomment this for manual test prototytping
# # (the test suite will run this if it's uncommented)
# def test_do_something(notebook_frontend):
#     # Do something with the notebook_frontend here
#     notebook_frontend.add_cell()
#     notebook_frontend.add_cell()
#     assert len(notebook_frontend.cells) == 3
#
#     notebook_frontend.delete_all_cells()
#     assert len(notebook_frontend.cells) == 1
#
#     notebook_frontend.editor_page.pause()
#     cell_texts = ['aa = 1', 'bb = 2', 'cc = 3']
#     a, b, c = cell_texts
#     notebook_frontend.populate(cell_texts)
#     assert notebook_frontend.get_cells_contents() == [a, b, c]
#     notebook_frontend._pause()



"""Test multiselect toggle

TODO: This changes the In []: label preceding the cell,
      what's the purpose of this? Update the docstring
"""


def test_prompt_numbers(prefill_notebook):
    notebook_frontend = prefill_notebook(['print("a")'])

    def get_prompt():
        return (
            notebook_frontend.cells[0].locate('.input')
            .locate('.input_prompt')
            .get_inner_html().strip()
        )

    def set_prompt(value):
        notebook_frontend.set_cell_input_prompt(0, value)

    assert get_prompt() == "<bdi>In</bdi>&nbsp;[&nbsp;]:"

    set_prompt(2)
    assert get_prompt() == "<bdi>In</bdi>&nbsp;[2]:"

    set_prompt(0)
    assert get_prompt() == "<bdi>In</bdi>&nbsp;[0]:"

    set_prompt("'*'")
    assert get_prompt() == "<bdi>In</bdi>&nbsp;[*]:"

    set_prompt("undefined")
    assert get_prompt() == "<bdi>In</bdi>&nbsp;[&nbsp;]:"

    set_prompt("null")
    assert get_prompt() == "<bdi>In</bdi>&nbsp;[&nbsp;]:"



"""Manager to read and modify config data in JSON files."""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import errno
import glob
import json
import os
import copy

from traitlets.config import LoggingConfigurable
from traitlets.traitlets import Unicode, Bool


def recursive_update(target, new):
    """Recursively update one dictionary using another.

    None values will delete their keys.
    """
    for k, v in new.items():
        if isinstance(v, dict):
            if k not in target:
                target[k] = {}
            recursive_update(target[k], v)
            if not target[k]:
                # Prune empty subdicts
                del target[k]

        elif v is None:
            target.pop(k, None)

        else:
            target[k] = v


def remove_defaults(data, defaults):
    """Recursively remove items from dict that are already in defaults"""
    # copy the iterator, since data will be modified
    for key, value in list(data.items()):
        if key in defaults:
            if isinstance(value, dict):
                remove_defaults(data[key], defaults[key])
                if not data[key]:  # prune empty subdicts
                    del data[key]
            else:
                if value == defaults[key]:
                    del data[key]


class BaseJSONConfigManager(LoggingConfigurable):
    """General JSON config manager

    Deals with persisting/storing config in a json file with optionally
    default values in a {section_name}.d directory.
    """

    config_dir = Unicode('.')
    read_directory = Bool(True)

    def ensure_config_dir_exists(self):
        """Will try to create the config_dir directory."""
        try:
            os.makedirs(self.config_dir, 0o755)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    def file_name(self, section_name):
        """Returns the json filename for the section_name: {config_dir}/{section_name}.json"""
        return os.path.join(self.config_dir, section_name+'.json')

    def directory(self, section_name):
        """Returns the directory name for the section name: {config_dir}/{section_name}.d"""
        return os.path.join(self.config_dir, section_name+'.d')

    def get(self, section_name, include_root=True):
        """Retrieve the config data for the specified section.

        Returns the data as a dictionary, or an empty dictionary if the file
        doesn't exist.

        When include_root is False, it will not read the root .json file,
        effectively returning the default values.
        """
        paths = [self.file_name(section_name)] if include_root else []
        if self.read_directory:
            pattern = os.path.join(self.directory(section_name), '*.json')
            # These json files should be processed first so that the
            # {section_name}.json take precedence.
            # The idea behind this is that installing a Python package may
            # put a json file somewhere in the a .d directory, while the
            # .json file is probably a user configuration.
            paths = sorted(glob.glob(pattern)) + paths
        self.log.debug('Paths used for configuration of %s: \n\t%s', section_name, '\n\t'.join(paths))
        data = {}
        for path in paths:
            if os.path.isfile(path):
                with open(path, encoding='utf-8') as f:
                    recursive_update(data, json.load(f))
        return data

    def set(self, section_name, data):
        """Store the given config data.
        """
        filename = self.file_name(section_name)
        self.ensure_config_dir_exists()

        if self.read_directory:
            # we will modify data in place, so make a copy
            data = copy.deepcopy(data)
            defaults = self.get(section_name, include_root=False)
            remove_defaults(data, defaults)

        # Generate the JSON up front, since it could raise an exception,
        # in order to avoid writing half-finished corrupted data to disk.
        json_content = json.dumps(data, indent=2)

        f = open(filename, 'w', encoding='utf-8')
        with f:
            f.write(json_content)

    def update(self, section_name, new_data):
        """Modify the config section by recursively updating it with new_data.

        Returns the modified config data as a dictionary.
        """
        data = self.get(section_name)
        recursive_update(data, new_data)
        self.set(section_name, data)
        return data



"""Tests shutdown of the Kernel."""
from .utils import EDITOR_PAGE

def test_shutdown(prefill_notebook):
    notebook_frontend = prefill_notebook(["print(21)"])
    
    notebook_frontend.try_click_selector('//a[text()="Kernel"]', page=EDITOR_PAGE)
    notebook_frontend.try_click_selector('#shutdown_kernel',  page=EDITOR_PAGE)
    notebook_frontend.try_click_selector('.btn.btn-default.btn-sm.btn-danger', page=EDITOR_PAGE)
    
    notebook_frontend.execute_cell(0)

    assert not notebook_frontend.is_kernel_running()
    assert not notebook_frontend.get_cell_output()



"""Test cell deletion"""


from .utils import TREE_PAGE, EDITOR_PAGE


def cell_is_deletable(notebook, index):
    JS = f'() => {{ return Jupyter.notebook.get_cell({index}).is_deletable(); }}'
    return notebook.evaluate(JS, EDITOR_PAGE)


def remove_all_cells(notebook):
    for i in range(len(notebook.cells)):
        notebook.delete_cell(0)


INITIAL_CELLS = ['print("a")', 'print("b")', 'print("c")']


def test_delete_cells(notebook_frontend):
    a, b, c = INITIAL_CELLS
    notebook_frontend.populate(INITIAL_CELLS)

    # Validate initial state
    assert notebook_frontend.get_cells_contents() == [a, b, c]
    for cell in range(0, 3):
        assert cell_is_deletable(notebook_frontend, cell)

    notebook_frontend.set_cell_metadata(0, 'deletable', 'false')
    notebook_frontend.set_cell_metadata(1, 'deletable', 0
    )
    assert not cell_is_deletable(notebook_frontend, 0)
    assert cell_is_deletable(notebook_frontend, 1)
    assert cell_is_deletable(notebook_frontend, 2)

    # Try to delete cell a (should not be deleted)
    notebook_frontend.delete_cell(0)
    assert notebook_frontend.get_cells_contents() == [a, b, c]

    # Try to delete cell b (should succeed)
    notebook_frontend.delete_cell(1)
    assert notebook_frontend.get_cells_contents() == [a, c]

    # Try to delete cell c (should succeed)
    notebook_frontend.delete_cell(1)
    assert notebook_frontend.get_cells_contents() == [a]

    # Change the deletable state of cell a
    notebook_frontend.set_cell_metadata(0, 'deletable', 'true')

    # Try to delete cell a (should succeed)
    notebook_frontend.delete_cell(0)
    assert len(notebook_frontend.cells) == 1 # it contains an empty cell

    # Make sure copied cells are deletable
    notebook_frontend.edit_cell(index=0, content=a)
    notebook_frontend.set_cell_metadata(0, 'deletable', 'false')
    assert not cell_is_deletable(notebook_frontend, 0)
    notebook_frontend.to_command_mode()
    notebook_frontend.type_active('cv')
    assert len(notebook_frontend.cells) == 2
    assert cell_is_deletable(notebook_frontend, 1)

    notebook_frontend.set_cell_metadata(0, 'deletable', 'true')  # to perform below test, remove all the cells
    notebook_frontend.delete_all_cells()
    assert len(notebook_frontend.cells) == 1    # notebook should create one automatically on empty notebook
